#include <stdio.h>
#include <unistd.h>
#include "common.h"
#include "board_mgr.h"

#define EQU3(a, b, c)   ((a) == (b) && (a) == (c) && (a) != EMPTY)

//forward declarations
static void destroy(board_mgr_t *self);
static int has_winner(char *board);
static int has_empty_spot(char *board);
static void print_board(char *mark, char *board);
static void read_board(int fd, char* board);
static void write_board(int fd, char* board);

//make a board manager
board_mgr_t *make_board_mgr() {
    board_mgr_t *bmgr = malloc(sizeof(board_mgr_t));
    ON_FALSE_EXIT(bmgr != NULL, "out of memory");

    bmgr->destroy        = destroy;
    bmgr->has_winner     = has_winner;
    bmgr->has_empty_spot = has_empty_spot;
    bmgr->print_board    = print_board;
    bmgr->read_board     = read_board;
    bmgr->write_board    = write_board;
    return bmgr;
}

//destroy the board manager
static void destroy(board_mgr_t *self) {
    free(self);
}

//whether there is a winner
static int has_winner(char *board) {
    char (*b)[3] = (char(*)[3])board;
    for(int i = 0; i < 3; i++) {
        if( EQU3(b[i][0], b[i][1], b[i][2]) ||
            EQU3(b[0][i], b[1][i], b[2][i]) )
            return 1;
    }

    return  EQU3(b[0][0], b[1][1], b[2][2]) ||
            EQU3(b[0][2], b[1][1], b[2][0]);
}

//whether there is an empty spot
static int has_empty_spot(char *board) {
    for(int i = 0; i < 9; i++)
        if(board[i] == EMPTY)
            return 1;
    return 0;
}

//print the board
static void print_board(char *mark, char *board) {
    printf("%s:\n", mark);
    printf("%c %c %c\n", board[0], board[1], board[2]);
    printf("%c %c %c\n", board[3], board[4], board[5]);
    printf("%c %c %c\n", board[6], board[7], board[8]);
}

//read the board from fd
static void read_board(int fd, char* board) {
    ON_FALSE_EXIT(read(fd, board, 9) == 9, "cannot read board");
}

//write the board to fd
static void write_board(int fd, char* board) {
    ON_FALSE_EXIT(write(fd, board, 9) == 9, "cannot write board");
}
