#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <unistd.h>
#include "common.h"
#include "board_mgr.h"

//Proxy for the actual palyer
//
typedef struct player_proxy {
    char *fname;        //executable file name for the player
    char *mark;         //mark of the player
    pid_t pid;          //the pid of the player
    int fd[2];          //pipe connected to this player
    board_mgr_t *bmgr;  //board manager

    void (*load       )(struct player_proxy *self);
    void (*end        )(struct player_proxy *self);
    void (*read_board )(struct player_proxy *self, char *board);
    void (*write_board)(struct player_proxy *self, char *board);
    void (*print_board)(struct player_proxy *self, char *board);
} player_proxy_t;

//The game
//
typedef struct game {
    //whose turn is it now (either 0 or 1)
    int turn;
    //proxy for the two players
    player_proxy_t *proxy[2];

    void (*load_players)(struct game *self);
    void (*end         )(struct game *self);
    void (*play        )(struct game *self);
} game_t;
game_t game;

//forward declarations
static player_proxy_t *make_player_proxy(char *cmd, char *mark);
static void player_load(player_proxy_t *self);
static void player_end(player_proxy_t *self);
static void player_read_board(player_proxy_t *self, char *board);
static void player_write_board(player_proxy_t *self, char *board);
static void player_print_board(player_proxy_t *self, char *board);

static game_t *make_game(char *player_pgm_1, char *player_pgm_2);
static void game_load_players(game_t *self);
static void game_end(game_t *self);
static void game_play(game_t *self);

// signal handler
//
volatile int switch_turn; 
static void sig_handler(int sig) {
    //TODO: if sig is SIGUSR1, set switch_turn to 1

}

/////////////////////////////////////////////////////////////////////
// player proxy
//

//make a player for the mark
static player_proxy_t *make_player_proxy(char *fname, char *mark) {
    player_proxy_t *proxy = malloc(sizeof(player_proxy_t));
    ON_FALSE_EXIT(proxy != NULL, "out of memory");

    proxy->load        = player_load;
    proxy->end         = player_end;
    proxy->read_board  = player_read_board;
    proxy->write_board = player_write_board;
    proxy->print_board = player_print_board;

    proxy->fname = fname;
    proxy->mark  = mark;
    proxy->bmgr  = make_board_mgr();

    //TODO: prepare proxy->fd using pipe

    return proxy;
}

//load the actual player
//   - the actual player is fname (an executable program)
//   - N.B. this method is called from a CHILD process
static void player_load(player_proxy_t *self) {
    //TODO: connect fd[0] to stdin and fd[1] to stdout using dup2,
    //      close unnecessary fds


    //TODO: execute fname using execvp
    //      - fname needs mark as its only parameter

}

//terminate the actual player and destroy the proxy
static void player_end(player_proxy_t *self) {
    //TODO: send SIGKILL to the actual player,
    //      reap the player,
    //      close unnecessary fds


    //TODO: destroy the board manager,
    //      free self

}

//read the board from the actual player
static void player_read_board(player_proxy_t *self, char *board) {
    //TODO: read the board from the actual player
    //hint: use bmgr->read_board

}

//write the baord to the actual player
static void player_write_board(player_proxy_t *self, char *board) {
    //TODO: write the board to the actual player
    //hint: use bmgr->write_board


    //TODO: send SIGUSR1 to the actual player

}

//print the board
static void player_print_board(player_proxy_t *self, char *board) {
    self->bmgr->print_board(self->mark, board);
}

/////////////////////////////////////////////////////////////////////
// game
//

//make a game
//   - player_pgm_*: executable file name for each player
static game_t *make_game(char *player_pgm_1, char *player_pgm_2) {
    game_t *game = malloc(sizeof(game_t));
    ON_FALSE_EXIT(game != NULL, "out of memory");

    game->load_players = game_load_players;
    game->end  = game_end;
    game->play = game_play;

    //TODO: make the proxy players of the game for marks "O" and "X"


    //TODO: set the turn of the game to index 0


    return game;
}

//load the players
static void game_load_players(game_t *self) {
    for(int i = 0; i < 2; i++) {
        player_proxy_t *player = self->proxy[i];

        //TODO: create a child process using fork
        //      - child: load the player
        //      - parent: set the player's pid

    }

    //wait for the players to be ready
    sleep(1);   
}

//end the game
static void game_end(game_t *self) {
    for(int i = 0; i < 2; i++) {
        player_proxy_t *player = self->proxy[i];
        //TODO: end the player

    }
    free(self);
}

//play the game
static void game_play(game_t *self) {
    char board[10] = ".........";

    //block SIGUSR1 signal using sigprocmask
    sigset_t mask, prev;
    //TODO: empty mask,
    //      add SIGUSR1 to mask,
    //      block signals in mask and get the previous set in prev


    //TODO: set sig_handler as a signal handler for SIGUSR1


    self->turn = 0;
    player_proxy_t *proxy = self->proxy[self->turn];
    switch_turn = 0;

    proxy->print_board(proxy, board);
    proxy->write_board(proxy, board);
    while(!proxy->bmgr->has_winner(board) &&
           proxy->bmgr->has_empty_spot(board)) {
               
        //wait until the player marks
        //TODO: until switch_turn is set, wait for a signal


        //read and print the board
        //TODO: read the board from the actual player using its proxy,
        //      print the board


        //check the result
        if(proxy->bmgr->has_winner(board)) {
            printf("winner: %s\n", proxy->mark);
            break;
        }
        if(!proxy->bmgr->has_empty_spot(board)) {
            printf("no winner: tie\n");
            break;
        }

        //switch the turn
        //TODO: switch the self->turn index (0 <-> 1),
        //      set the proxy variable,
        //      set switch_turn to 0


        //let the other player play
        //TODO: write the board to the actual player using its proxy

    }
}

int main(int argc, char **argv) {
    ON_FALSE_EXIT(argc == 3,
        strmsg("usage: %s player_1_cmd player_2_cmd", argv[0]));

    game_t *game = make_game(argv[1], argv[2]);
    game->load_players(game);
    game->play(game);
    game->end(game);

    return 0;
}
