#include <signal.h>
#include <unistd.h>
#include "common.h"
#include "board_mgr.h"

// Actual player program
//
typedef struct player {
    char *mark;         //either "O" or "X"
    board_mgr_t *bmgr;  //board manager

    void (*destroy      )(struct player *self);
    void (*read_board   )(struct player *self, char *board);
    void (*write_board  )(struct player *self, char *board);
    int  (*simulate_play)(struct player *self, char mark, char *board, int *inx);
    int  (*put_mark     )(struct player *self, char *board);
    void (*play         )(struct player *self);
} player_t;

//forward delcarations
static player_t *make_player(char *mark);
static void destroy(player_t *self);
static void player_read_board(player_t *self, char *board);
static void player_write_board(player_t *self, char *board);
static int player_simulate_play(player_t *self, char mark, char *board, int *inx);
static int  player_put_mark(player_t *self, char *board);
static void player_play(player_t *self);

// signal handler
//
static volatile int my_turn = 0;
static void sig_handler(int sig) {
    //TODO: if sig is SIGUSR1, set my_turn to 1

}

/////////////////////////////////////////////////////////////////////
// player
//

//make a player for the mark
static player_t *make_player(char *mark) {
    player_t *player = malloc(sizeof(player_t));
    ON_FALSE_EXIT(player != NULL, "out of memory");

    player->read_board    = player_read_board;
    player->write_board   = player_write_board;
    player->simulate_play = player_simulate_play;
    player->put_mark      = player_put_mark;
    player->play          = player_play;

    player->mark = mark;
    player->bmgr = make_board_mgr();

    return player;
}

//destroy the player
static void destroy(player_t *self) {
    self->bmgr->destroy(self->bmgr);
    free(self);
}

//read board from STDIN
static void player_read_board(player_t *self, char *board) {
    //TODO: read board from STDIN_FILENO
    //hint: use bmgr->read_board

}

//write board to STDOUT
static void player_write_board(player_t *self, char *board) {
    //TODO: write board to STDOUT_FILENO
    //hint: use bmgr->write_board


    //TODO: send SIGUSR1 signal to the game
    //hint: game is the parent of this process

}

//simulated play for mark
static int player_simulate_play(player_t *self, char mark, char *board, int *inx) {
    UNUSED(self);
    int i, win = 0;
    for(i = 0; i < 9 && !win; i++) {
        if(board[i] == EMPTY) {
            board[i] = mark;
            win = self->bmgr->has_winner(board);
            board[i] = EMPTY;
            *inx = i; // any empty place
        }
    }
    return win;
}

//put mark on the board
static int player_put_mark(player_t *self, char *board) {
    int i = -1;
    if(!player_simulate_play(self, OTHER(self->mark[0]), board, &i))
        player_simulate_play(self, self->mark[0], board, &i);
    if(i >= 0)
        board[i] = self->mark[0];
    return i >= 0;
}

//play the game
static void player_play(player_t *self) {
    char board[10] = ".........";

    //block SIGUSR1 signal using sigprocmask
    sigset_t mask, prev;
    //TODO: empty mask,
    //      add SIGUSR1 to mask,
    //      block signals in mask and get the previous set in prev


    //TODO: set sig_handler as a signal handler for SIGUSR1


    while(!self->bmgr->has_winner(board) &&
           self->bmgr->has_empty_spot(board)) {

        //wait for my turn
        //TODO: until my_turn is set, wait for a signal
        //hint: use sigsuspend


        //set my turn to 0
        my_turn = 0;

        //read board
        //TODO: read the board from stdin (actually from a proxy player)


        if(self->bmgr->has_winner(board))
            break;
        if(!self->bmgr->has_empty_spot(board))
            break;

        //mark on the board
        self->put_mark(self, board);

        //write board
        //TODO: write the board to stdout (actually to a proxy player)

    }
}

int main(int argc, char **argv) {
    ON_FALSE_EXIT(argc == 2, strmsg("usage: %s [X | O]", argv[0]));

    player_t *player = make_player(argv[1]);
    player->play(player);
    player->destroy(player);

    return 0;
}
